# 空结构体和空接口


## 空结构体

### 定义

空结构体是指在Go语言中定义的一个没有任何字段的结构体类型。在Go语言中，结构体是一种复合数据类型，可以包含零个或多个字段，每个字段可以是不同的数据类型。

空结构体的定义有两种方式，方式1：

```go
type EmptyStruct struct{}
var est EmptyStruct
```

方式2：

```go
var est struct{}
```

空结构体的特点：

1. **零内存占用**：空结构体不包含任何字段，因此它在内存中不占用任何空间。这意味着使用空结构体作为数据类型可以在一些场景中节省内存。
2. **类型唯一性**：每个空结构体类型都是唯一的，即使是在不同的包中定义的空结构体，只要它们的定义是相同的，那么它们仍然是相同的类型。

空结构体是一种特殊的类型，其实例也只能有唯一值，也就是`struct{}{}`，当我们定义了一个空结构体类型的变量时，它的默认值就是`struct{}{}`，我们也不能对其进行修改。

### 零内存占用

尽管空结构体在Go语言中不占用内存，但编译器会确保每个空结构体实例在内存中有一个唯一的地址。这样做是为了确保每个空结构体实例都有一个唯一的标识符，以便在程序中能够准确地区分它们。

当我们创建一个空结构体实例时，编译器会为其**分配一个唯一的地址**，但由于空结构体本身没有任何字段，因此不需要额外的内存来存储实际的数据。因此，尽管每个空结构体实例都有一个地址，但它们的实际占用的内存空间是零。

> 编译器为每个变量分配的地址只是一个标识符，用于在程序执行期间识别和访问变量。这个地址并不存储任何实际的数据，仅用于标识变量在内存中的位置。

这也意味着，尽管我们可以创建无数个空结构体实例，但它们之间不会占用额外的内存空间，因为它们共享相同的类型信息，并且每个实例都指向相同的零字节内存。

示例：

```go
type EST struct{}

func main() {
	var est1 EST
	fmt.Printf("%p\n", &est1) // 0xcc62e0

	var est2 struct{}
	var est3 struct{}
	fmt.Printf("%p\n", &est2) // 0xcc62e0
	fmt.Printf("%p\n", &est3) // 0xcc62e0
}
```

上面的例子中，所有空结构体类型的变量的地址都相同，说明他们是共用的同一个地址标识符。

### 使用场景

这里介绍两种空结构体的使用场景。

#### 通道中用作信号传递

```go
// 空结构体作为通道元素类型，用于表示事件的发生
type Event struct{}

// 通道用于通知事件发生
var eventCh chan Event

// 模拟事件的产生
func produceEvents() {
    for i := 0; i < 3; i++ {
       eventCh <- Event{}
    }
    close(eventCh)
}

// 模拟事件的处理
func consumeEvents() {
    for {
       _, ok := <-eventCh
       if !ok {
          break
       }
       fmt.Println("Event occurred!")
    }
}

func main() {
    eventCh = make(chan Event)
    go produceEvents()
    consumeEvents()
}
```

在这个示例中，空结构体`Event`用作通道元素的类型，表示某种事件的发生。当通道接收到`Event{}`时，即表示事件发生。

#### 实现集合中的集合

```go
// 空结构体作为集合元素类型
type Set map[string]struct{}

func main() {
    set := make(Set)
    set["apple"] = struct{}{}
    set["banana"] = struct{}{}
    set["orange"] = struct{}{}

    // 检查元素是否存在于集合中
    fmt.Println("Is apple in set?", isInSet(set, "apple"))
    fmt.Println("Is pear in set?", isInSet(set, "pear"))
}

// 检查元素是否存在于集合中
func isInSet(set Set, item string) bool {
    _, found := set[item]
    return found
}
```

在这个示例中，`Set`类型是一个映射，其中键是字符串，值是空结构体。这个结构实现了一个简单的集合，用于存储字符串元素。由于空结构体不占用任何内存空间，因此它适合用作集合中的元素类型，因为它不需要额外的内存消耗。

### struct{}和struct{}{}的区别

#### struct{}

- `struct{}`是一个空的结构体类型声明。
- 这种声明可以被用于声明一个变量、作为参数类型、或者作为返回类型。

#### struct{}{}

- `struct{}{}`是一个空结构体字面量，表示创建一个空的结构体实例。
- 这种字面量通常被用作创建空结构体实例，比如在通道通信中用作通道元素发送或接收。

总的来说，`struct{}`用于声明空的结构体类型，而`struct{}{}`用于创建一个具体的空结构体实例。

## 空接口

在 Go 语言中，空接口 `interface{}` 是一种特殊类型的接口，也被称为**空接口类型**或**万能类型**。它没有任何方法，因此可以表示任意类型的值。

> 空接口可以表示任意类型的值，因为任何类型都至少实现了零个方法。

空接口在 Go 中的主要作用是提供一种通用的方式来处理各种类型的数据，类似于其他语言中的泛型类型或者 `Object` 类型。

以下是空接口的一些重要特性和用法：

1. **任意类型的容器**：空接口可以容纳任意类型的值。这使得可以在不指定具体类型的情况下，存储和传递各种不同类型的数据。

2. **类型断言**：要在运行时获取空接口中的具体类型，需要使用类型断言。这允许你检查空接口中存储的值的实际类型，并以适当的方式处理它。

   ```go
   var val interface{} = 42
   intValue, ok := val.(int)
   if ok {
       fmt.Println("Value is an integer:", intValue)
   } else {
       fmt.Println("Value is not an integer")
   }
   ```

3. **与空切片一起使用**：空接口经常与空切片结合使用，以实现通用的数据集合。通过将各种类型的数据存储在空接口的切片中，可以创建通用的容器，其中可以容纳任何类型的数据。

   ```go
   var data []interface{}
   data = append(data, 42, "hello", 3.14)
   ```

4. **函数参数和返回值**：空接口经常用作函数的参数类型或返回类型，以实现接受或返回任意类型的值的函数。

   ```go
   func process(val interface{}) {
       // 处理函数
   }
   ```

5. **空接口的零值**：空接口的零值为 `nil`。当空接口没有指向任何值时，它的值为 `nil`。

空接口在 Go 中是非常灵活且强大的，但在使用时需要注意一些潜在的陷阱，如类型安全和性能开销。使用空接口时，应该谨慎考虑类型转换和类型断言，并确保在可能的情况下尽量避免使用空接口来维持代码的可读性和性能。

---

> 作者: [Hollis](https://www.xuexiqu.cn/)  
> URL: https://www.xuexiqu.cn/golang/interview/empty_structure_and_empty_interface/  

